from collections.abc import Callable, Iterable
from typing import Any

from koza.transform import KozaTransform, Record

Tag = str | list[str] | None


class KozaTransformHook:
    def __init__(self, fn: Callable[..., Any], tag: Tag):
        self.fn = fn
        self.tag = tag


# @koza.prepare_data()
# Pre-process data before continuing with transform
class KozaPrepareDataFunction(KozaTransformHook):
    def __call__(self, koza: KozaTransform, data: Iterable[Record]) -> Iterable:
        return self.fn(koza, data)


def prepare_data(tag: Tag = None):
    """
    Decorator to prepare input data before starting a transform.

    The return value of a function marked with this decorator will be used as
    the input for the transform function(s) for this transform.

    Usage:

        @koza.prepare_data()
        def convert_to_pandas(koza: KozaTransform, data: Iterable[dict[str, Any]]):
            return pd.DataFrame(data)

        @koza.transform()
        def transform_with_pandas(koza: KozaTransform: data: pd.DataFrame):
            # `data` here is the pandas data frame returned from above

    Any sort of iterator may be returned from this function. For example, a cursor
    from a database query:

        @koza.prepare_data()
        def load_db_data(koza, data):
            conn = db.connect("localhost:9152")
            res = conn.query("SELECT * FROM table")
            return res

        @koza.transform_record()
        def transform_db_row(koza, record):
            # `record` here will be a row from the database query

    :param tag: The tag with which this hook should be associated.
    """

    def decorator(fn: Callable[[KozaTransform, Record], Iterable | None]):
        return KozaPrepareDataFunction(fn, tag)

    return decorator


# @koza.transform()
# Mark a function as being a single transform
class KozaSingleTransformFunction(KozaTransformHook):
    def __call__(self, koza: KozaTransform, data: Iterable[Record]) -> Iterable | None:
        return self.fn(koza, data)


def transform(tag: Tag = None):
    """
    Decorator to mark a transformation function.

    A function marked with this hook will be run *once* for all input data for a
    given tag. If your transform can be performed on records one at a time,
    prefer to use `@koza.transform_record()`.

    Usage:

        @koza.transform()
        def do_transform(koza: KozaTransform, data):
            for record in data:
                output = MyOutputObject(
                    name=record["name"],
                    label=record["label"],
                )

                koza.write(output)

    :param tag: The tag with which this hook should be associated.
    """

    def decorator(fn: Callable[[KozaTransform, Iterable[Record]], Iterable | None]):
        return KozaSingleTransformFunction(fn, tag)

    return decorator


# @koza.transform_record()
# Mark a function as being a function to transform single records
class KozaSerialTransformFunction(KozaTransformHook):
    def __call__(self, koza: KozaTransform, data: Record) -> Iterable | None:
        return self.fn(koza, data)


def transform_record(tag: Tag = None):
    """
    Decorator to mark a record transformation function.

    This function will be called on every record for a configured reader. By
    default, these records will be of the type `dict[str, Any]`, but may be
    different if a `@koza.prepare_data` decorator has been configured.

    Usage:

        @koza.transform_record()
        def transform_csv_row(koza: KozaTransform, record: dict[str, Any]):
            output = MyOutputObject(
                name=record["name"],
                label=record["label"],
            )

            koza.write(output)

    :param tag: The tag with which this hook should be associated.
    """

    def decorator(fn: Callable[[KozaTransform, Record], Iterable | None]):
        return KozaSerialTransformFunction(fn, tag)

    return decorator


# @koza.on_data_begin()
# Mark a function as being called before the reader starts
class KozaDataBeginFunction(KozaTransformHook):
    def __call__(self, koza: KozaTransform):
        self.fn(koza)


def on_data_begin(tag: Tag = None):
    """
    Decorator to mark a function to be called before data is read in a transform.

    :param tag: The tag with which this hook should be associated.
    """

    def decorator(fn: Callable[[KozaTransform], None]):
        return KozaDataBeginFunction(fn, tag)

    return decorator


# @koza.on_read_end
# Mark a function as being called after the reader ends
class KozaDataEndFunction(KozaTransformHook):
    def __call__(self, koza: KozaTransform):
        self.fn(koza)


def on_data_end(tag: Tag = None):
    """
    Decorator to mark a function to be called after data is read in a transform.

    :param tag: The tag with which this hook should be associated.
    """

    def decorator(fn: Callable[[KozaTransform], None]):
        return KozaDataEndFunction(fn, tag)

    return decorator



import koza


@koza.transform_record()
def transform_record(koza: koza.KozaTransform, record: dict):
    koza.write(
        {
            "STRING": record["STRING"],
            "entrez": record["entrez"],
        }
    )



import pytest
from biolink_model.datamodel.pydanticmodel_v2 import Gene, GeneToGeneAssociation  # , Publication

from koza.converter.kgx_converter import KGXConverter


def test_gene_conversion():
    fgf8a = Gene(
        id="ZFIN:ZDB-GENE-990415-72",
        symbol="fgf8a",
        name="fibroblast growth factor 8a",
        in_taxon=["NCBITaxon:7955"],
    )

    kgx_converter: KGXConverter = KGXConverter()
    output = kgx_converter.convert_node(fgf8a)

    assert output["id"] == "ZFIN:ZDB-GENE-990415-72"
    assert output["symbol"] == "fgf8a"
    assert "gene" in str(output["category"]).lower()
    assert output["name"] == "fibroblast growth factor 8a"
    assert output["in_taxon"] == ["NCBITaxon:7955"]


def test_association_conversion():
    fgf8a = Gene(id="ZFIN:ZDB-GENE-990415-72", symbol="fgf8a", name="fibroblast growth factor 8a")
    pax2a = Gene(id="ZFIN:ZDB-GENE-990415-8", symbol="pax2a", name="paired box 2a")
    pub = "PMID:17522161"  # Publication(id="PMID:17522161", type="MESH:foobar")
    association = GeneToGeneAssociation(
        id="uuid:123",
        subject=fgf8a.id,
        predicate="biolink:interacts_with",
        object=pax2a.id,
        publications=[pub],
        knowledge_level="not_provided",
        agent_type="not_provided",
    )

    kgx_converter: KGXConverter = KGXConverter()
    (_, edges) = kgx_converter.split_entities([fgf8a, pax2a, association])

    output = kgx_converter.convert_association(edges[0])
    assert output["subject"] == "ZFIN:ZDB-GENE-990415-72"
    assert output["object"] == "ZFIN:ZDB-GENE-990415-8"
    # TODO figure out how/where to handle this conversion
    # assert Curie("PMID:17522161") in output['publications']


@pytest.mark.parametrize(
    "id, symbol, synonym, xref",
    [
        ("MGI:1917258", "Ace2", ["RIKEN cDNA 2010305L05 gene", "2010305L05Rik"], []),
        ("RGD:620474", "Sox9", [], ["ENSEMBL:ENSRNOG00000002607", "NCBI_Gene:140586"]),
    ],
)
def test_keys_uniformity(id, symbol, synonym, xref):
    """
    Connfirm that the result of the conversion has the same fields, even if some are empty lists
    """
    gene = Gene(id=id, symbol=symbol, synonym=synonym, xref=xref)

    output = KGXConverter().convert_node(gene)

    assert "category" in output.keys()
    assert "id" in output.keys()
    assert "symbol" in output.keys()
    assert "synonym" in output.keys()
    assert "xref" in output.keys()
    assert "description" not in output.keys()


@pytest.mark.parametrize(
    "id, symbol, synonym, xref",
    [
        ("MGI:1917258", None, None, []),
        ("RGD:620474", "", [], ["ENSEMBL:ENSRNOG00000002607", "NCBI_Gene:140586"]),
    ],
)
def test_exclude_none(id, symbol, synonym, xref):
    """
    Connfirm that the result of the conversion has the same fields, even if some are empty lists
    """
    gene = Gene(id=id, symbol=symbol, synonym=synonym, xref=xref)

    output = KGXConverter().convert_node(gene)

    assert output["id"] == "MGI:1917258" or output["id"] == "RGD:620474"

    if output["id"] == "MGI:1917258":
        assert "symbol" not in output.keys()
        assert "synonym" not in output.keys()
    elif output["id"] == "RGD:620474":
        assert "symbol" in output.keys()
        assert "synonym" in output.keys()

    assert "category" in output.keys()
    assert "id" in output.keys()
    assert "description" not in output.keys()


import re
import uuid

from biolink_model.datamodel.pydanticmodel_v2 import PairwiseGeneToGeneInteraction, Protein
from koza.model.graphs import KnowledgeGraph

import koza


@koza.transform()
def string_transform(koza: koza.KozaTransform, data):
    for row in data:
        protein_a = Protein(id="ENSEMBL:" + re.sub(r"\d+\.", "", row["protein1"]))
        protein_b = Protein(id="ENSEMBL:" + re.sub(r"\d+\.", "", row["protein2"]))

        pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
            id="uuid:" + str(uuid.uuid1()),
            subject=protein_a.id,
            object=protein_b.id,
            predicate="biolink:interacts_with",
            knowledge_level="not_provided",
            agent_type="not_provided",
        )

        yield KnowledgeGraph(nodes=[protein_a, protein_b], edges=[pairwise_gene_to_gene_interaction])



"""
Testing the biolink config dataclasses + pydandic

"""

import pytest
from pydantic import TypeAdapter, ValidationError

from koza.model.formats import OutputFormat
from koza.model.koza import KozaConfig
from koza.model.reader import ReaderConfig


@pytest.mark.parametrize(
    "inclusion, column, filter_code, value",
    (
        [
            ("include", "combined_score", "lt", "70"),
            ("exclude", "combined_score", "lt", "70"),
            ("include", "combined_score", "in", "70"),
            ("exclude", "combined_score", "in", "70"),
            ("exclude", "combined_score", "in", 70),
            ("exclude", "combined_score", "in", 0.7),
            ("include", "combined_score", "eq", ["goat", "sheep"]),
            ("include", "combined_score", "lt", ["goat", "sheep"]),
            ("include", "combined_score", "ge", ["goat", "sheep"]),
        ]
    ),
)
def test_wrong_filter_type_raises_exception(inclusion, column, filter_code, value):
    """
    Test if include and exclude raise a
    value error when handed an incompatible type,
    eg a string when using the lt operator
    """
    config = {
        "filters": [
            {
                "column": column,
                "inclusion": inclusion,
                "filter_code": filter_code,
                "value": value,
            }
        ],
    }
    with pytest.raises(ValidationError) as e:
        TypeAdapter(ReaderConfig).validate_python(config)

    for error in e.value.errors():
        assert error["msg"].startswith("Input should be a")


@pytest.mark.parametrize(
    "inclusion, code",
    (
        [
            ("include", "lgt"),
            ("exclude", "ngte"),
        ]
    ),
)
def test_wrong_filter_code_raises_exception(inclusion, code):
    config = {
        "filters": [
            {
                "inclusion": inclusion,
                "filter_code": code,
            }
        ],
    }
    with pytest.raises(ValidationError) as e:
        TypeAdapter(ReaderConfig).validate_python(config)

    assert e.value.error_count() == 1
    assert e.value.errors()[0]["msg"].startswith(
        f"Input tag '{code}' found using 'filter_code' does not match any of the expected tags:"
    )


def test_filter_on_nonexistent_column():
    config = {
        "name": "test_config",
        "reader": {
            "columns": ["a", "b", "c"],
            "filters": [
                {
                    "column": "a",
                    "inclusion": "include",
                    "filter_code": "gt",
                    "value": 0,
                },
                {
                    "column": "d",
                    "inclusion": "include",
                    "filter_code": "gt",
                    "value": 0,
                },
                {
                    "column": "e",
                    "inclusion": "include",
                    "filter_code": "gt",
                    "value": 0,
                },
            ],
        },
    }

    with pytest.raises(ValidationError) as e:
        TypeAdapter(KozaConfig).validate_python(config)

    assert e.value.error_count() == 1
    assert e.value.errors()[0]["msg"].startswith(
        f"Value error, One or more filter columns not present in designated CSV columns: 'd', 'e'"
    )


def test_tagged_readers():
    config_dict = {
        "name": "test_config",
        "readers": {
            "tag1": {
                "format": "jsonl",
            },
            "tag2": {
                "format": "csv",
            },
        },
    }

    config = TypeAdapter(KozaConfig).validate_python(config_dict)
    readers = config.get_readers()
    assert len(readers) == 2
    assert readers[0].tag == "tag1"
    assert readers[1].tag == "tag2"
    assert readers[0].reader.format == OutputFormat.jsonl


def test_file_archive_in_new_config():
    """Test that file_archive is supported in the new config format"""
    config_dict = {
        "name": "test_config",
        "reader": {
            "format": "csv",
            "files": ["file1.csv"],
            "file_archive": "archive.tar.gz",
        },
    }

    config = TypeAdapter(KozaConfig).validate_python(config_dict)
    assert config.reader is not None
    assert config.reader.file_archive == "archive.tar.gz"
    assert "file1.csv" in config.reader.files
